/*
   base64.cpp and base64.h

   base64 encoding and decoding with C++.

   Version: 1.01.00

   Copyright (C) 2004-2017 René Nyffenegger

   This source code is provided 'as-is', without any express or implied
   warranty. In no event will the author be held liable for any damages
   arising from the use of this software.

   Permission is granted to anyone to use this software for any purpose,
   including commercial applications, and to alter it and redistribute it
   freely, subject to the following restrictions:

   1. The origin of this source code must not be misrepresented; you must not
      claim that you wrote the original source code. If you use this source code
      in a product, an acknowledgment in the product documentation would be
      appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and must not be
      misrepresented as being the original source code.

   3. This notice may not be removed or altered from any source distribution.

   René Nyffenegger rene.nyffenegger@adp-gmbh.ch

*/

#include "arrow/util/base64.h"
#include <array>
#include <cstdint>
#include <iostream>

namespace arrow {
namespace util {

constexpr std::string_view base64_chars =
             "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
             "abcdefghijklmnopqrstuvwxyz"
             "0123456789+/";

static const std::array<int8_t, 256> kBase64Lookup = [] {
  std::array<int8_t, 256> table{};
  table.fill(-1);

  for (size_t i = 0; i < base64_chars.size(); ++i) {
    table[static_cast<uint8_t>(base64_chars[i])] = static_cast<int8_t>(i);
  }

  return table;
}();

static std::string base64_encode(unsigned char const* bytes_to_encode, unsigned int in_len) {
  std::string ret;
  int i = 0;
  int j = 0;
  unsigned char char_array_3[3];
  unsigned char char_array_4[4];

  while (in_len--) {
    char_array_3[i++] = *(bytes_to_encode++);
    if (i == 3) {
      char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
      char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
      char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
      char_array_4[3] = char_array_3[2] & 0x3f;

      for(i = 0; (i <4) ; i++)
        ret += base64_chars[char_array_4[i]];
      i = 0;
    }
  }

  if (i)
  {
    for(j = i; j < 3; j++)
      char_array_3[j] = '\0';

    char_array_4[0] = ( char_array_3[0] & 0xfc) >> 2;
    char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
    char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);

    for (j = 0; (j < i + 1); j++)
      ret += base64_chars[char_array_4[j]];

    while((i++ < 3))
      ret += '=';

  }

  return ret;

}

std::string base64_encode(std::string_view string_to_encode) {
  auto bytes_to_encode = reinterpret_cast<const unsigned char*>(string_to_encode.data());
  auto in_len = static_cast<unsigned int>(string_to_encode.size());
  return base64_encode(bytes_to_encode, in_len);
}

Result<std::string> base64_decode(std::string_view encoded_string) {
  size_t in_len = encoded_string.size();
  int i = 0;
  std::string_view::size_type in_ = 0;
  int padding_count = 0;
  int block_padding = 0;
  bool padding_started = false;
  unsigned char char_array_4[4], char_array_3[3];
  std::string ret;

  if (encoded_string.size() % 4 != 0) {
    return Status::Invalid("Invalid base64 input: length is not a multiple of 4");
  }

  while (in_len--) {
    unsigned char c = encoded_string[in_];

    if (c == '=') {
      padding_started = true;
      padding_count++;

      if (padding_count > 2) {
        return Status::Invalid("Invalid base64 input: too many padding characters");
      }

      char_array_4[i++] = 0;
    } else {
      if (padding_started) {
        return Status::Invalid("Invalid base64 input: padding in wrong position");
      }

      int8_t val = kBase64Lookup[static_cast<uint8_t>(c)];

      if (val == -1) {
        return Status::Invalid("Invalid base64 input: character is not valid base64 character");
      }

      char_array_4[i++] = val;
    }

    in_++;

    if (i == 4) {
      char_array_3[0] = (char_array_4[0] << 2) + ((char_array_4[1] & 0x30) >> 4);
      char_array_3[1] = ((char_array_4[1] & 0xf) << 4) + ((char_array_4[2] & 0x3c) >> 2);
      char_array_3[2] = ((char_array_4[2] & 0x3) << 6) + char_array_4[3];

      block_padding = padding_count;

      for (i = 0; i < 3 - block_padding; i++) {
        ret += char_array_3[i];
      }

      i = 0;
    }
  }

  return ret;
}

}  // namespace util
}  // namespace arrow
