// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

#pragma once

#include <cstdint>
#include <span>

#include "parquet/bloom_filter.h"

namespace parquet::internal {

inline bool FindHashBlockImpl(
    std::span<const uint32_t, BlockSplitBloomFilter::kBitsSetPerBlock> block,
    std::span<const uint32_t, BlockSplitBloomFilter::kBitsSetPerBlock> salt,
    uint32_t key) {
  uint32_t miss = 0;
  for (int i = 0; i < BlockSplitBloomFilter::kBitsSetPerBlock; ++i) {
    const uint32_t mask = static_cast<uint32_t>(1) << ((key * salt[i]) >> 27);
    miss |= (~block[i] & mask);
  }
  return miss == 0;
}

}  // namespace parquet::internal
